from threading import Thread
from queue import Queue
from socket import socket
from logging import getLogger
from uuid import UUID

from edri.config.constant import SWITCH_BYTES_LENGTH


class Sender(Thread):
    def __init__(self, connection_queue: Queue[bytes], connection: socket, router_id: UUID) -> None:
        super(Sender, self).__init__()
        self.connection_queue = connection_queue
        self.connection = connection
        self.logger = getLogger(f"{__name__}.{router_id}")
        self.router_id = router_id

    def run(self) -> None:
        while True:
            message = self.connection_queue.get()
            self.logger.debug("sending %s",  len(message))
            message_size = len(message)
            try:
                self.connection.send(message_size.to_bytes(SWITCH_BYTES_LENGTH, "big", signed=False) + message)
            except BrokenPipeError:
                self.connection_queue.put(message)
                return
            except Exception:
                self.connection_queue.put(message)
                return
            self.logger.debug("send %s",  len(message))



from http import HTTPStatus
from json import dumps, JSONEncoder
from typing import Callable, Unpack, TypedDict, Type

from edri.api import Headers
from edri.api.handlers import HTTPHandler
from edri.api.handlers.http_handler import ResponseErrorKW
from edri.config.constant import ApiType
from edri.dataclass.directive import HTTPResponseDirective
from edri.dataclass.event import Event
from edri.utility import NormalizedDefaultDict


class ResponseKW(TypedDict):
    headers: NormalizedDefaultDict[str, Headers]


class RESTHandler[T: HTTPResponseDirective](HTTPHandler):
    _directive_handlers: dict[T, Callable[[T], tuple[int, Headers]]] = {}

    def __init__(self, scope: dict, receive: Callable, send: Callable, headers: NormalizedDefaultDict[str, Headers], json_encoder: Type[JSONEncoder]):
        super().__init__(scope, receive, send, headers)
        self.json_encoder = json_encoder

    @classmethod
    def api_type(cls) -> ApiType:
        return ApiType.REST

    async def response(self, status: HTTPStatus, data: Event | bytes, *args, **kwargs: Unpack[ResponseKW]):
        headers = kwargs["headers"]
        if isinstance(data, Event):
            response = data.get_response()
            context = data.get_context()
            if "Content-Type" not in headers:
                headers["Content-Type"].append("application/json")
            try:
                body = dumps(response.as_dict(transform=True, keep_concealed=False), ensure_ascii=False, cls=self.json_encoder, context=context).encode("utf-8")
            except TypeError as e:
                self.logger.error("Object is not JSON serializable", exc_info=e)
                await self.response_error(HTTPStatus.INTERNAL_SERVER_ERROR, {
                    "reasons": [{
                        "status_code": HTTPStatus.INTERNAL_SERVER_ERROR,
                        "message": "Object is not JSON serializable",
                    }]
                })
                return
        else:
            body = data

        return await super().response(status, body, headers=headers)

    async def response_error(self, status: HTTPStatus, data: Event | dict | None = None, *args, **kwargs: Unpack[ResponseErrorKW]):
        data = self.response_error_prepare(status, data)
        data = dumps(data, ensure_ascii=False, cls=self.json_encoder).encode("utf-8")
        if not "headers" in kwargs:
            kwargs["headers"] = NormalizedDefaultDict[str, Headers](list)
        kwargs["headers"]["Content-Type"].append("application/json")
        return await super().response_error(status, data, *args, **kwargs)



from datetime import datetime, timedelta
from typing import Optional

from edri.dataclass.event import Event, event
from edri.events.edri.group import Scheduler


@event
class Update(Scheduler):
    identifier: str
    event: Optional[Event] = None
    when: Optional[datetime] = None
    repeat: Optional[timedelta] = None



from typing import Optional
from uuid import UUID

from edri.dataclass.event import Event, event
from edri.dataclass.response import Response, response
from edri.events.edri.group import Router


@response
class SendFromResponse(Response):
    event: Event


@event
class SendFrom(Router):
    router_id: UUID
    key: Optional[str] = None
    response: SendFromResponse


from dataclasses import dataclass
from inspect import getfile
from logging import Logger, getLogger
from multiprocessing import Queue
from threading import Timer
from typing import Type

from watchdog.events import FileSystemEventHandler, FileModifiedEvent, DirModifiedEvent
from watchdog.observers import Observer

from edri.abstract import ManagerBase
from edri.config.setting import ENVIRONMENT
from edri.events.edri.manager import Restart


@dataclass
class WatcherHandlerComponent:
    type: Type[ManagerBase]
    timer: Timer | None = None


class WatcherHandler(FileSystemEventHandler):
    def __init__(self, components: dict[str, Type[ManagerBase]], router_queue: Queue, logger: Logger):
        self.router_queue = router_queue
        self.components: dict[str, WatcherHandlerComponent] = {path: WatcherHandlerComponent(component) for path, component in
                                                               components.items()}
        self.logger = logger
        self.delay = 1 if ENVIRONMENT == "development" else 10

    def on_modified(self, event: FileModifiedEvent | DirModifiedEvent):
        if isinstance(event, FileModifiedEvent):
            component = self.components[event.src_path]
            self.logger.debug("Manager %s was changed", component.type.__name__)
            if component.timer is not None:
                self.logger.debug("Resetting timer: %s", component.type.__name__)
                component.timer.cancel()
            component.timer = Timer(self.delay, self._send_restart_event, args=(component.type,))
            component.timer.start()

    def _send_restart_event(self, manager_type: Type[ManagerBase]):
        self.logger.info("Restart event queued for manager: %s", manager_type.__name__)
        restart = Restart(manager=manager_type)
        self.router_queue.put(restart)

    def quit(self):
        for component in self.components.values():
            if component.timer:
                component.timer.cancel()


class Watcher:
    def __init__(self, router_queue: Queue, components: set[ManagerBase]) -> None:
        self.router_queue = router_queue
        self.logger = getLogger(__name__)
        self.observer = Observer()
        components = {getfile(component.__class__): component.__class__ for component in components}
        handler = WatcherHandler(components, self.router_queue, self.logger)
        for path in components.keys():
            self.observer.schedule(handler, path)
        self.observer.start()

    def quit(self) -> None:
        self.observer.stop()



from dataclasses import dataclass
from datetime import datetime
from typing import Optional

from edri.dataclass.directive import HTTPResponseDirective


@dataclass
class CookieResponseDirective(HTTPResponseDirective):
    """
    Represents a directive for setting an HTTP cookie in a response.

    Attributes:
        name (str): The name of the cookie.
        value (str): The value of the cookie.
        expires (Optional[datetime]): The date and time when the cookie expires. If not set, the cookie will be a session cookie.
        path (Optional[str]): The path within the domain where the cookie is valid. Defaults to the root path ('/').
        comment (Optional[str]): A comment for the cookie, providing additional information to the client. This is rarely used.
        domain (Optional[str]): The domain where the cookie is valid. If not set, defaults to the host of the request URL.
        max_age (Optional[int]): The maximum age of the cookie in seconds. If set, it overrides the `expires` attribute.
        secure (bool): Indicates if the cookie should only be transmitted over secure protocols like HTTPS.
        version (Optional[int]): The version of the cookie. Version 0 is the default and refers to "old" cookies, while version 1 is the newer cookie format.
        httponly (bool): Indicates if the cookie is accessible only through HTTP(S) and not available to JavaScript (helps mitigate cross-site scripting attacks).
        samesite (Optional[str]): Controls whether a cookie is sent with cross-site requests. Possible values are 'Lax', 'Strict', or 'None'.
    """
    name: str
    value: str
    expires: Optional[datetime] = None
    path: Optional[str] = None
    comment: Optional[str] = None
    domain: Optional[str] = None
    max_age: Optional[int] = None
    secure: bool = False
    version: Optional[int] = None
    httponly: bool = False
    samesite: Optional[str] = None


@dataclass
class AccessDeniedResponseDirective(HTTPResponseDirective):
    message: str | None = None


@dataclass
class UnauthorizedResponseDirective(HTTPResponseDirective):
    message: str | None = None


@dataclass
class NotFoundResponseDirective(HTTPResponseDirective):
    message: str | None = None

@dataclass
class ConflictResponseDirective(HTTPResponseDirective):
    message: str | None = None

@dataclass
class HeaderResponseDirective(HTTPResponseDirective):
    name: str
    value: str

@dataclass
class UnprocessableContentResponseDirective(HTTPResponseDirective):
    message: str | None = None
